/*
 * kinectserver - a program that send data from the Kinect sensor to connected clients.
 * Copyright (c) 2011, University of Kent.
 * Contact: Neil Brown, nccb@kent.ac.uk
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <boost/smart_ptr.hpp>
#include <wx/app.h>
#include <wx/frame.h>
#include <wx/icon.h>
#include <wx/menu.h>
#include <wx/string.h>
#include <wx/taskbar.h>
#include <wx/textctrl.h>
#include <wx/timer.h>

#include "server.h"

#include "greenfoot.xpm"
#include "greenfoot-stopped.xpm"

class LogFrame : public wxFrame
{
  wxTextCtrl* text;
  Log log;
  LogFrame** parentPointer;
public:
  inline LogFrame(Log logParam, LogFrame** pp)
    : wxFrame(NULL, -1, wxString("Greenfoot Kinect Log", wxConvUTF8)), log(logParam), parentPointer(pp)
  {
    text = new wxTextCtrl(this, -1, wxString(), wxDefaultPosition, wxDefaultSize, wxTE_MULTILINE | wxTE_READONLY);
  }

  void update()
  {
    wxString s(log.getText().c_str(), wxConvUTF8);
    if (s.Len() != text->GetValue().Len())
      text->SetValue(s);
  }

  void close(wxCloseEvent& e)
  {
    //Hide on close, if we can:
    if (e.CanVeto())
    {
      e.Veto();
      Show(false);
    }
    else
    {
      Destroy();
      *parentPointer = NULL;
    }
  }

  DECLARE_EVENT_TABLE()
};

BEGIN_EVENT_TABLE(LogFrame, wxFrame)
  EVT_CLOSE(LogFrame::close)
END_EVENT_TABLE()

class TrayApp;

class TrayMenu : public wxMenu
{
  TrayApp* app;
public:
  TrayMenu(TrayApp* appParam);

  void menuExit(wxCommandEvent&);
  void menuStart(wxCommandEvent&);
  void menuStop(wxCommandEvent&);
  void menuRestart(wxCommandEvent&);
  void menuLog(wxCommandEvent&);


  DECLARE_EVENT_TABLE()
};

BEGIN_EVENT_TABLE(TrayMenu, wxMenu)
  EVT_MENU(wxID_EXIT, TrayMenu::menuExit)
  EVT_MENU(1, TrayMenu::menuStart)
  EVT_MENU(2, TrayMenu::menuStop)
  EVT_MENU(3, TrayMenu::menuRestart)
  EVT_MENU(4, TrayMenu::menuLog)
END_EVENT_TABLE()

class TrayIcon : public wxTaskBarIcon
{
  TrayApp* app;
  bool normal;
public:
  TrayIcon(TrayApp* appParam)
    : app(appParam)
  {
    SetIcon(wxIcon(greenfoot_xpm), wxString("Greenfoot Kinect Server", wxConvUTF8));
    normal = true;
  }

  void setNormalIcon()
  {
    if (normal) return;
    SetIcon(wxIcon(greenfoot_xpm), wxString("Greenfoot Kinect Server", wxConvUTF8));
    normal = true;
  }

  void setStoppedIcon()
  {
    if (false == normal) return;
    SetIcon(wxIcon(greenfoot_stopped_xpm), wxString("Greenfoot Kinect Server", wxConvUTF8));
    normal = false;
  }

  wxMenu* CreatePopupMenu()
  {
    return new TrayMenu(app);
  }
};

bool setup(int argc, wchar_t** argv, Log log, Server* server, boost::scoped_ptr<Time>* idleTimeout);

class TrayApp : public wxApp
{
  Logger log;
  Server server;
  TrayIcon* icon;
  LogFrame* logFrame;
  bool status; // true if generating successfully
  boost::scoped_ptr<Time> idleTimeout;
  wxTimer tim;
public:
  TrayApp() : server(log), icon(NULL), logFrame(NULL), status(true), tim(this, 99)
  {
  }

  bool OnInit()
  {
    if (false == setup(argc, argv, log, &server, &idleTimeout))
    {
      log.printStdout();
      return false;
    }
    logFrame = new LogFrame(log, &logFrame);
    icon = new TrayIcon(this);
    tim.Start(1000, wxTIMER_CONTINUOUS); //events every second
    return icon->IsOk();
  }

  void OnIdle(wxIdleEvent& e)
  {
    status = server.process(idleTimeout.get(), true);
    if (logFrame != NULL) logFrame->update();
    if (status)
    {
      icon->setNormalIcon();
      e.RequestMore(true);
    }
    else 
    {
      icon->setStoppedIcon();
      // Don't request more idle, just wait for timer
    }
  }

  void OnTimer(wxTimerEvent& e)
  {
    //We need to set a timer to make sure idle events happen every once in a while
    // for when a client reconnects.
    //Otherwise nothing will happen until they triggered a GUI event.

    wxWakeUpIdle();
  }

  int OnExit()
  {
    delete icon;
    return 0;
  }

  friend class TrayMenu;

  DECLARE_EVENT_TABLE()
};

BEGIN_EVENT_TABLE(TrayApp, wxApp)
  EVT_IDLE(TrayApp::OnIdle)
  EVT_TIMER(99, TrayApp::OnTimer)
END_EVENT_TABLE()

IMPLEMENT_APP(TrayApp)

void TrayMenu::menuExit(wxCommandEvent&)
{
  app->Exit();
}

void TrayMenu::menuStart(wxCommandEvent&)
{
  app->status = app->server.start();
}

void TrayMenu::menuStop(wxCommandEvent&)
{
  app->server.stop();
}

void TrayMenu::menuRestart(wxCommandEvent& e)
{
  menuStop(e);
  menuStart(e);
}

void TrayMenu::menuLog(wxCommandEvent&)
{
  if (app->logFrame != NULL)
    app->logFrame->Show();
}


TrayMenu::TrayMenu(TrayApp* appParam)
    : app(appParam)
{
  if (app->status)
  {
    Append(2, wxString("Stop", wxConvUTF8));
    Append(3, wxString("Restart", wxConvUTF8));
  }
  else
    Append(1, wxString("Start", wxConvUTF8));
  Append(4, wxString("Show Log", wxConvUTF8));
  Append(wxID_EXIT, wxString("Exit", wxConvUTF8));
}
