/*
 * kinectserver - a program that send data from the Kinect sensor to connected clients.
 * Copyright (c) 2011, University of Kent.
 * Some of this code is copied from an OpenNI sample, Copyright (c) PrimeSense 2010.
 * Contact: Neil Brown, nccb@kent.ac.uk
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "interface.h"
#include "mytime.h"
#include <XnTypes.h>
#include <XnOpenNI.h>
#include <XnCppWrapper.h>

#include <map>

#define CHECK_RC(nRetVal, what)\
	if (nRetVal != XN_STATUS_OK)\
	{\
          sprintf(errorBuf, "%s failed: %s\n", what, xnGetStatusString(nRetVal));\
		return errorBuf;\
	}

unsigned gtmp;

#define CHECK(op)\
  if (XN_STATUS_OK != (gtmp = op)) { printf("Error code: %d\n", gtmp); return 0;}

const char* defaultConfig =
  "<OpenNI><Licenses></Licenses>"
  "<Log writeToConsole=\"true\" writeToFile=\"false\">"
    "<LogLevel value=\"2\"/><Masks><Mask name=\"ALL\" on=\"true\"/></Masks><Dumps></Dumps>"
  "</Log>"
  "<ProductionNodes>"
    "<Node type=\"Image\" name=\"Image1\"><Configuration>"
       "<MapOutputMode xRes=\"640\" yRes=\"480\" FPS=\"30\"/><Mirror on=\"true\"/>"
    "</Configuration></Node>"
    "<Node type=\"Depth\" name=\"Depth1\"><Configuration>"
       "<MapOutputMode xRes=\"640\" yRes=\"480\" FPS=\"30\"/><Mirror on=\"true\"/>"
    "</Configuration></Node>"
  "</ProductionNodes>"
  "</OpenNI>"
  ;


//////////////////////////////////////////////////////////////////////
// KinectImpl declaration:
//////////////////////////////////////////////////////////////////////

class KinectImpl
{
public:
  xn::Context context;
  xn::UserGenerator userGenerator;
  xn::DepthGenerator depthGenerator;
  xn::ImageGenerator imageGenerator;
  xn::ImageMetaData imageMD;
  xn::DepthMetaData depthMD;
  XnBool needPose;
  char strPose[20];
  std::map<XnUserID, User> users;
  char errorBuf[1024];
  int pixels[640*480];
  Time lastUpdateFinished;
  bool showTimes;
  bool generating;
  Log log;

  KinectImpl(Log logParam)
    : log(logParam)
  {
    needPose = FALSE;
    showTimes = false;
    generating = false;
    strcpy(strPose, "");
  }

  inline const char* init(const char* xmlPath);
  inline bool start();
  inline void stop();
  inline bool updateJointInfo();
  inline bool getUsers(User* userArray);
  inline Point3 worldToScreen(const Point3&);
  inline const unsigned char* getRGB();
  inline const unsigned short* getDepth();
  inline const unsigned short* getPixelLabels();
  inline InterfaceStatus process();
  inline bool matchDepthRGB();
  inline unsigned short getMaxDepth();
  inline std::pair<int, int> getDepthDim();
};

//////////////////////////////////////////////////////////////////////
// KinectInterface implementation:
//////////////////////////////////////////////////////////////////////

KinectInterface::KinectInterface(Log log)
  : impl(new KinectImpl(log))
{
}

KinectInterface::~KinectInterface()
{
  delete impl;
}

const char* KinectInterface::init(const char* xmlPath)
{
  return impl->init(xmlPath);
}

void KinectInterface::stop()
{
  impl->stop();
}

bool KinectInterface::start()
{
  return impl->start();
}

InterfaceStatus KinectInterface::process()
{
  return impl->process();
}

int KinectInterface::numUsers() const
{
  return impl->users.size();
}

bool KinectInterface::getUsers(User* userArray) const
{
  return impl->getUsers(userArray);
}

const unsigned char* KinectInterface::getRGB() const
{
  return impl->getRGB();
}

const unsigned short* KinectInterface::getDepth() const
{
  return impl->getDepth();
}

const unsigned short* KinectInterface::getUserLabels() const
{
  return impl->getPixelLabels();
}

void KinectInterface::showTimings()
{
  impl->showTimes = true;
  impl->lastUpdateFinished.setCurrentTime();
}

bool KinectInterface::matchDepthRGB()
{
  return impl->matchDepthRGB();
}

unsigned short KinectInterface::getMaxDepth()
{
  return impl->getMaxDepth();
}

std::pair<int, int> KinectInterface::getDepthDim()
{
  return impl->getDepthDim();
}

//////////////////////////////////////////////////////////////////////
// KinectImpl callbacks:
//////////////////////////////////////////////////////////////////////


void XN_CALLBACK_TYPE UserCalibration_CalibrationStart(xn::SkeletonCapability& capability, XnUserID nId, void* pCookie)
{
  KinectImpl* i = (KinectImpl*)pCookie;
  i->log << "Begun calibrating user " << nId << "\n";
  i->users[nId].id = nId;
  i->users[nId].state = STATE_CALIBRATING;
}

void XN_CALLBACK_TYPE User_LostUser(xn::UserGenerator& generator, XnUserID nId, void* pCookie)
{
  KinectImpl* i = (KinectImpl*)pCookie;
  i->log << "Lost user " << nId << "\n";
  i->users.erase(nId);
}

void XN_CALLBACK_TYPE User_NewUser(xn::UserGenerator& generator, XnUserID nId, void* pCookie)
{
  KinectImpl* i = (KinectImpl*)pCookie;
  i->log << "New user " << nId << "\n";
  // New user found                                                                                                                                        
  if (i->needPose)
    {
      i->userGenerator.GetPoseDetectionCap().StartPoseDetection(i->strPose, nId);
      i->users[nId].id = nId;
      i->users[nId].state = STATE_DETECTING_POSE;
    }
  else
    {
      i->userGenerator.GetSkeletonCap().RequestCalibration(nId, TRUE);
      i->users[nId].id = nId;
      i->users[nId].state = STATE_CALIBRATING;
    }
}

void XN_CALLBACK_TYPE UserCalibration_CalibrationEnd(xn::SkeletonCapability& capability, XnUserID nId, XnBool bSuccess, void* pCookie)
{
  KinectImpl* i = (KinectImpl*)pCookie;
  if (bSuccess)
    {
      // Calibration succeeded                                                                                                                         
      i->log << "Calibration complete, start tracking user " << nId << "\n";
      i->userGenerator.GetSkeletonCap().StartTracking(nId);
      i->users[nId].id = nId;
      i->users[nId].state = STATE_TRACKING;
    }
  else
    {
      // Calibration failed                                                                                                                            
      User_NewUser(i->userGenerator, nId, pCookie);
    }
}

//////////////////////////////////////////////////////////////////////
// KinectImpl implementation:
//////////////////////////////////////////////////////////////////////


void XN_CALLBACK_TYPE UserPose_PoseDetected(xn::PoseDetectionCapability& capability, const XnChar* strPose, XnUserID nId, void* pCookie)
{
  KinectImpl* i = (KinectImpl*)pCookie;
  i->userGenerator.GetPoseDetectionCap().StopPoseDetection(nId);
  i->userGenerator.GetSkeletonCap().RequestCalibration(nId, TRUE);
  i->users[nId].id = nId;
  i->users[nId].state = STATE_CALIBRATING;
}

const char* KinectImpl::init(const char* xmlPath)
{
	XnStatus nRetVal = XN_STATUS_OK;

		xn::EnumerationErrors errors;
		nRetVal = context.InitFromXmlFile(xmlPath, &errors);
		if (nRetVal == XN_STATUS_NO_NODE_PRESENT)
		{
			XnChar strError[1024];
			errors.ToString(strError, 1024);
			sprintf(errorBuf,"%s\n", strError);
			return errorBuf;
		}
		else if (nRetVal != XN_STATUS_OK)
		{
                  sprintf(errorBuf, "Open failed: %s\n", xnGetStatusString(nRetVal));
			return errorBuf;
		}
	
                xn::NodeInfoList list;
                nRetVal = context.EnumerateExistingNodes(list);
                CHECK_RC(nRetVal, "Find image node")
                  for (xn::NodeInfoList::Iterator it = list.Begin(); it != list.End(); ++it)
                      {
                        if ((*it).GetDescription().Type == XN_NODE_TYPE_IMAGE)
                          (*it).GetInstance(imageGenerator);
                      }


	nRetVal = context.FindExistingNode(XN_NODE_TYPE_DEPTH, depthGenerator);
	CHECK_RC(nRetVal, "Find depth generator");
	nRetVal = context.FindExistingNode(XN_NODE_TYPE_USER, userGenerator);
	if (nRetVal != XN_STATUS_OK)
	{
		nRetVal = userGenerator.Create(context);
		CHECK_RC(nRetVal, "Find user generator");
	}

	XnCallbackHandle hUserCallbacks, hCalibrationCallbacks, hPoseCallbacks;
	if (!userGenerator.IsCapabilitySupported(XN_CAPABILITY_SKELETON))
	{
          sprintf(errorBuf, "Supplied user generator doesn't support skeleton\n");
		return errorBuf;
	}
	userGenerator.RegisterUserCallbacks(User_NewUser, User_LostUser, this, hUserCallbacks);
	userGenerator.GetSkeletonCap().RegisterCalibrationCallbacks(UserCalibration_CalibrationStart, UserCalibration_CalibrationEnd, this, hCalibrationCallbacks);

	if (userGenerator.GetSkeletonCap().NeedPoseForCalibration())
	{
		needPose = TRUE;
		if (!userGenerator.IsCapabilitySupported(XN_CAPABILITY_POSE_DETECTION))
		{
                  sprintf(errorBuf,"Pose required, but not supported\n");
			return errorBuf;
		}
		userGenerator.GetPoseDetectionCap().RegisterToPoseCallbacks(UserPose_PoseDetected, NULL, this, hPoseCallbacks);
		userGenerator.GetSkeletonCap().GetCalibrationPose(strPose);
	}

	userGenerator.GetSkeletonCap().SetSkeletonProfile(XN_SKEL_PROFILE_ALL);

	nRetVal = context.StartGeneratingAll();
	CHECK_RC(nRetVal, "StartGenerating");
        generating = true;

        strcpy(errorBuf, "");
        return errorBuf;
}

void KinectImpl::stop()
{
  context.StopGeneratingAll();
  generating = false;
}

bool KinectImpl::start()
{
  generating = (XN_STATUS_OK == context.StartGeneratingAll());
  return generating;
}

bool KinectImpl::matchDepthRGB()
{
  return XN_STATUS_OK == depthGenerator.GetAlternativeViewPointCap().SetViewPoint(imageGenerator);
}

const static XnSkeletonJoint joints[] = {
  XN_SKEL_HEAD,           //0
  XN_SKEL_NECK,           //1
  XN_SKEL_TORSO,          //2
  XN_SKEL_LEFT_SHOULDER,  //3
  XN_SKEL_LEFT_ELBOW,     //4
  XN_SKEL_LEFT_HAND,      //5
  XN_SKEL_RIGHT_SHOULDER, //6
  XN_SKEL_RIGHT_ELBOW,    //7
  XN_SKEL_RIGHT_HAND,     //8
  XN_SKEL_LEFT_HIP,       //9
  XN_SKEL_LEFT_KNEE,      //10
  XN_SKEL_LEFT_FOOT,      //11
  XN_SKEL_RIGHT_HIP,      //12
  XN_SKEL_RIGHT_KNEE,     //13
  XN_SKEL_RIGHT_FOOT};    //14

const static std::pair<int, int> jointConns[] = {
  std::make_pair(0, 1),   //Head->Neck

  std::make_pair(1, 3),   //Neck->LShoulder
  std::make_pair(3, 4),   //LShoulder->LElbow
  std::make_pair(4, 5),   //LElbow->LHand

  std::make_pair(1, 6),   //Neck->RShoulder
  std::make_pair(6, 7),   //RShoulder->RElbow
  std::make_pair(7, 8),   //RElbow->RHand

  std::make_pair(3, 2),   //LShoulder->Torso
  std::make_pair(6, 2),   //RShoulder->Torso

  std::make_pair(2, 9),   //Torso->LHip
  std::make_pair(9, 10),  //LHip->LKnee
  std::make_pair(10, 11), //LKnee->LFoot

  std::make_pair(2, 12),   //Torso->RHip
  std::make_pair(12, 13),  //RHip->RKnee
  std::make_pair(13, 14),  //RKnee->RFoot
  };

std::vector< std::pair<int, int> > KinectInterface::jointConnections()
{
  return std::vector< std::pair<int, int> >(&(jointConns[0]), &(jointConns[14]));
}

inline void copyPoint3(Point3& dest, const XnPoint3D& src)
{
  dest.x = src.X;
  dest.y = src.Y;
  dest.z = src.Z;
}

bool KinectImpl::updateJointInfo()
  {
    XnSkeletonJointPosition jp;
    for (std::map<XnUserID, User>::iterator it = users.begin(); it != users.end(); it++)
    {
      for (int i = 0;i < NUM_JOINTS;i++)
        {
          CHECK(userGenerator.GetSkeletonCap().GetSkeletonJointPosition(it->first, joints[i], jp));
          it->second.joints[i].confidence = jp.fConfidence;
          copyPoint3(it->second.joints[i].positionWorld, jp.position);
          it->second.joints[i].positionScreen = worldToScreen(it->second.joints[i].positionWorld);
        }
      XnPoint3D com;
      CHECK(userGenerator.GetCoM(it->first,com));
      copyPoint3(it->second.centreOfMassWorld, com);
      it->second.centreOfMassScreen = worldToScreen(it->second.centreOfMassWorld);
    }

    return true;
  }

bool KinectImpl::getUsers(User* userArray)
{
  if (false == updateJointInfo())
    return false;
  int i = 0;
  for (std::map<XnUserID, User>::iterator it = users.begin(); it != users.end(); it++)
    {
      userArray[i] = it->second;
      i += 1;
    }
  return true;
}

const unsigned short* KinectImpl::getPixelLabels()
{
  xn::SceneMetaData smd;
  CHECK(userGenerator.GetUserPixels(0, smd));
  return smd.Data();
}

const unsigned char* KinectImpl::getRGB()
{
  if (imageGenerator.IsValid())
    {
      imageGenerator.GetMetaData(imageMD);
      return imageMD.Data();
    }
 else
   return NULL;
}

const unsigned short* KinectImpl::getDepth()
{
  if (depthGenerator.IsValid())
    {
      return depthGenerator.GetDepthMap();
    }
  else
    return NULL;
}

unsigned short KinectImpl::getMaxDepth()
{
  if (depthGenerator.IsValid())
    return depthGenerator.GetDeviceMaxDepth();
  else
    return 0;
}

std::pair<int, int> KinectImpl::getDepthDim()
{
  xn::DepthMetaData depthMD;
  if (depthGenerator.IsValid())
    {
      depthGenerator.GetMetaData(depthMD);
      return std::make_pair(depthMD.XRes(), depthMD.YRes());
    }
  else
    return std::make_pair(-1, -1);
}

Point3 KinectImpl::worldToScreen(const Point3& src)
{
  XnPoint3D in, out;
  in.X = src.x;
  in.Y = src.y;
  in.Z = src.z;
  depthGenerator.ConvertRealWorldToProjective(1, &in, &out);
  Point3 r;
  r.x = out.X;
  r.y = out.Y;
  r.z = out.Z;
  return r;
}

InterfaceStatus KinectImpl::process()
{
  if (false == generating)
    return InterfaceNotGenerating;

  Time t, diff, proc;
  if (showTimes)
    {
      t.setCurrentTime();
      diff = t - lastUpdateFinished;
    }
  // Actual processing of sensor:
  unsigned r = context.WaitAndUpdateAll();
  if (r != XN_STATUS_OK)
  {
    log << "Problem with interface, error code: " << r << "\n";
    return InterfaceError;
  }
  if (showTimes)
    {
      lastUpdateFinished.setCurrentTime();
      proc = lastUpdateFinished - t;
      log << "Time in user code: "
          << diff.getSeconds()
          << " secs, time updating sensor: "
          << proc.getSeconds()
          << "secs\n";
    }
  return InterfaceOK;
}

bool writeToTempFile(const char* text, Log log, char* fileName)
{
#ifdef WIN32
  char tempPath[MAX_PATH];
  if (0 == GetTempPathA(MAX_PATH, tempPath))
    return false;
  if (0 == GetTempFileNameA(tempPath, "GKT", 0, fileName))
    return false;
  HANDLE f = CreateFileA(fileName, GENERIC_WRITE, 0, NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
  if (INVALID_HANDLE_VALUE == f)
    return false;
  DWORD written;
  if (FALSE == WriteFile(f, text, strlen(text), &written, NULL) || written != strlen(text))
    return false;
  CloseHandle(f);
  return true;
#else
  strcpy(fileName, "/tmp/XXXXXX.xml");
  int f = mkstemps(fileName, 4);
  if (f == -1)
    {
      log << "Could not make temporary file\n";
      return false;
    }
  const int len = strlen(text);
  if (len != write(f, text, len))
    {
      log << "Error writing to temporary file: " << fileName << "\n";
      return false;
    }
  close(f);
  return true;
#endif
}


KinectInterface* KinectInterfaceFactory::openInterface()
{
  char configPath[1024];
  if (configFile == "")
  {
    if (false == writeToTempFile(defaultConfig, log, configPath))
    {
      log << "Error writing config to temporary file\n";
      return NULL;
    }
  }
  else
  {
    strcpy(configPath, configFile.c_str());
  }

  KinectInterface* kinect = new KinectInterface(log);
  log << "Initialising Kinect sensor...\n";
  const char* res = kinect->init(configPath);

  if (strcmp("", res) == 0)
  {
    log << "Initialised Kinect sensor successfully\n";
  }
  else
  {
    log << "Initialisation of Kinect sensor failed: " << res << "\n";
    delete kinect;
    return NULL;
  }


  if (matchDepthRGB)
    kinect->matchDepthRGB();
  if (showTimings)
    kinect->showTimings();    

  return kinect;
}
