/*
 * kinectserver - a program that send data from the Kinect sensor to connected clients.
 * Copyright (c) 2011, University of Kent.
 * Contact: Neil Brown, nccb@kent.ac.uk
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

// All the data structures needed for getting data from the Kinect
#ifndef INTERFACE_H
#define INTERFACE_H

#include "log.h"
#include <utility>
#include <vector>

struct Point3
{
  float x;
  float y;
  float z;
};

struct Joint
{
  Point3 positionWorld;
  Point3 positionScreen;
  float confidence;
};

#define NUM_JOINTS 15

struct User
{
  int id;
  int state;
  Joint joints[NUM_JOINTS];
  Point3 centreOfMassWorld;
  Point3 centreOfMassScreen;
};

#define STATE_DETECTING_POSE 0
#define STATE_CALIBRATING 1
#define STATE_TRACKING 2

class KinectImpl;

enum InterfaceStatus
{
  InterfaceOK,
  InterfaceNotGenerating,
  InterfaceError
};

class KinectInterface : public boost::noncopyable
{
 private:
  KinectImpl* const impl;
  KinectInterface(Log log);
 public:
  ~KinectInterface();

  /// non-blank if there is a problem.
  /// Also effectively invokes start, so there is no need to call start directly afterwards.
  const char* init(const char* xmlPath);

  void stop();
  bool start();

  /// false if there is a problem, or if generation is currently stopped.
  InterfaceStatus process();
  /// -1 if there is a problem
  int numUsers() const;
  /// false if there is a problem
  bool getUsers(User* userArray) const;
  Point3 worldToScreen(Point3) const;
  /// NULL if there is a problem
  const unsigned char* getRGB() const;
  /// NULL if there is a problem
  const unsigned short* getDepth() const;
  // 640*480 array of unsigned short.
  // The ones that belong to a user are labelled with their id
  /// NULL if there is a problem
  const unsigned short* getUserLabels() const;
  void showTimings();
  /// False if there is a problem
  bool matchDepthRGB();
  /// 0 if there is a problem
  unsigned short getMaxDepth();

  /// std::make_pair(-1, -1) if there is a problem
  std::pair<int, int> getDepthDim();

  static std::vector< std::pair<int, int> > jointConnections();

  friend class KinectInterfaceFactory;
};

class KinectInterfaceFactory : public boost::noncopyable
{
private:
  Log log;
  std::string configFile;
  bool matchDepthRGB;
  bool showTimings;
public:
 KinectInterfaceFactory(Log _log, std::string _configFile, bool _matchDepthRGB, bool _showTimings)
   : log(_log), configFile(_configFile), matchDepthRGB(_matchDepthRGB), showTimings(_showTimings)
  {
  }

  KinectInterface* openInterface();
};

#endif
