/*
 * kinectserver - a program that send data from the Kinect sensor to connected clients.
 * Copyright (c) 2011, University of Kent.
 * Contact: Neil Brown, nccb@kent.ac.uk
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <boost/asio.hpp>
#include "interface.h"
#include "server.h"

#include <string.h>
#include <errno.h>

#include <boost/thread.hpp>
#include <boost/smart_ptr.hpp>

#include <vector>
#include <fstream>

using namespace boost::asio;
using namespace boost::asio::ip;


//////////////////////////////////////////////////////////////////////
// ServerImpl declaration:
//////////////////////////////////////////////////////////////////////

struct Data
{
  // Outdated, removed:
  // REQ:
  // 4-bytes; value: 0)
  // 4-bytes; unsigned int: requested width)
  // 4-bytes; unsigned int: requested height)
  // RESP:
  // 4-bytes; unsigned int: message size (not including these 4 bytes, or thumbnail at end)
  // 4-bytes; unsigned int: number of users
  // Per user:
  //   4-bytes; unsigned int: user id
  //   4-bytes; unsigned int: state
  //   <SCAN-LINES>
  //   If state is 2:
  //     <JOINT-DATA>
  // width*height*4 bytes: ARGB (MSB->LSB) image array (not stored in version0data)
  //std::vector<unsigned char> version0data; 

  // REQ:
  // 4-bytes; value: 1)
  // 4-bytes; unsigned int: requested width)
  // 4-bytes; unsigned int: requested height)
  // RESP:
  // 4-bytes; unsigned int: message size (not including these 4 bytes, or labels, depth or thumbnail at end)
  // 4-bytes; unsigned int: number of users
  // Per user:
  //   4-bytes; unsigned int: user id
  //   4-bytes; unsigned int: state
  //   If state is 2:
  //     <JOINT-DATA>
  // <USER-LABELS> (not stored in version1data)
  // width*height*4 bytes: ARGB (MSB->LSB) image array (not stored in version1data)
  //
  // USER-LABELS:
  // 4-bytes; unsigned int: total size
  // 2-bytes per label, if highest bit is 1 then it's a replication count for the following value.
  // Replication may cross line borders.  Total size: width * height
  std::vector<unsigned char> version1data;
  // If REQ begins with a 2 instead of 1, version1data is sent,
  // followed by 2 bytes for max depth, followed by 640*480*2 bytes for depth

  boost::array<unsigned short, 640*480 > fullLabelArray; //2-bytes per pixel, user id

  unsigned short maxDepth;

  boost::array<unsigned short, 640*480 > fullDepthArray; //2-bytes per pixel, depth value

  boost::array<unsigned char, 640*480*3 > fullImageArray; // 3-bytes per pixel, R, G, B
};

class ServerImpl
{
private:
  int refCount;
  friend inline void intrusive_ptr_add_ref(ServerImpl* p)
  {
    p->refCount += 1;
  }
  friend inline void intrusive_ptr_release(ServerImpl* p)
  {
    p->refCount -= 1;
    if (p->refCount <= 0)
      delete p;
  }
private:
  boost::asio::io_service ioService;

  boost::mutex dataMutex;
  boost::shared_ptr< Data > sharedData;

  boost::condition_variable dataNotifyCond;
  boost::mutex dataNotifyMutex;
  unsigned dataNotifyFrame;


  int clientCount;
  Time lastClientDisconnect;
  boost::mutex clientInfoMutex;

  KinectInterfaceFactory* kinectFactory;
  KinectInterface* kinect;
  Log log;
  std::string feedDir;

  bool getData(Data* r);

  static void runClientThread(boost::intrusive_ptr<ServerImpl> p, tcp::socket* s, int id) { p->handleConnection(s, id); }
  static void runAcceptThread(boost::intrusive_ptr<ServerImpl> p, tcp::acceptor* acc) { p->acceptClients(acc); }
  static void runFeedThread(boost::intrusive_ptr<ServerImpl> p) { p->storeFeed(); }
  void handleConnection(tcp::socket*, int);
  void acceptClients(tcp::acceptor*);
  void storeFeed();

  /// count and delta will be set on return, may be NULL
  void timeSinceLastClient(int* count, Time* delta);
public:
  inline ServerImpl(Log logParam) : refCount(0), dataNotifyFrame(0), log(logParam) {}

  bool init(KinectInterfaceFactory* p, std::string feedDir, bool network);
  bool start();
  void stop();
  bool process(const Time* idle, bool attemptRestartIfClients, int maxRetries);
};

//////////////////////////////////////////////////////////////////////
// Server implementation:
//////////////////////////////////////////////////////////////////////

Server::Server(Log log)
{
  impl = new ServerImpl(log);
}

Server::~Server()
{
}

bool Server::init(KinectInterfaceFactory* p, std::string feedDir, bool network)
{
  return impl->init(p, feedDir, network);
}

bool Server::process(const Time* idle, bool attemptRestart, int maxRetries)
{
  return impl->process(idle, attemptRestart, maxRetries);
}

bool Server::start()
{
  return impl->start();
}

void Server::stop()
{
  return impl->stop();
}


//////////////////////////////////////////////////////////////////////
// ServerImpl implementation:
//////////////////////////////////////////////////////////////////////

bool ServerImpl::init(KinectInterfaceFactory* p, std::string _feedDir, bool network)
{
  kinectFactory = p;
  feedDir = _feedDir;
  kinect = kinectFactory->openInterface();

  clientCount = 0;
  lastClientDisconnect.setCurrentTime();

  try
  {
    tcp::acceptor* acc = new tcp::acceptor(ioService, network ? tcp::endpoint(tcp::v4(), 0x2020) : tcp::endpoint(address_v4::loopback(), 0x2020), true);
    boost::thread serverThread(runAcceptThread, this, acc);
    if (feedDir != "")
    {
      boost::thread feedThread(runFeedThread, this);
    }
    return true;
  }
  catch(...) //TODO catch properly and print error
  {
    log << "Exception thrown while starting server";
    return false;
  }
  
}

bool ServerImpl::start()
{
  //An explicit start resets the time since last client disconnect:
  clientInfoMutex.lock();
  lastClientDisconnect.setCurrentTime();
  clientInfoMutex.unlock();

  return kinect->start();
}

void ServerImpl::stop()
{
  kinect->stop();
}

bool ServerImpl::process(const Time* idle, bool attemptRestartIfClients, int maxRetries)
{
  boost::shared_ptr< Data > newData(new Data);

  InterfaceStatus state;  

  if (kinect == NULL)
    state = InterfaceError;
  else
    state = kinect->process();

  switch (state)
  {
  case InterfaceOK:
    if (getData(newData.get()))
    {
      //Put it in to class while holding lock:
      dataMutex.lock();
      sharedData = newData;
      dataMutex.unlock();

      // Wake up all blocked socket handling threads:
      dataNotifyMutex.lock();
      dataNotifyFrame += 1;
      dataNotifyMutex.unlock();
      dataNotifyCond.notify_all();

      if (idle != NULL)
      {
          int clientCount;
          Time delta;
          timeSinceLastClient(&clientCount, &delta);
          if (clientCount == 0 && *idle < delta)
          {
              kinect->stop();
              return false;
          }
      }

      return true;
    }
    // Otherwise, fall through into the not generating case:
  case InterfaceNotGenerating:
    {
      int clientCount;
      timeSinceLastClient(&clientCount, NULL);

      if (attemptRestartIfClients && clientCount > 0 && maxRetries >= 1)
      {
        start();
        return process(idle, false, maxRetries - 1);
      }
      else
        return false;
    }
    break;
  case InterfaceError:
    // We always restart after an error, even if we're idle:
    delete kinect;
    kinect = kinectFactory->openInterface();
    if (maxRetries >= 1)
      return process(idle, attemptRestartIfClients, maxRetries - 1);
    else
      return false;
  default:
    return false; // Shouldn't happen
  }
}

void ServerImpl::timeSinceLastClient(int* count, Time* delta)
{
  Time t;
  t.setCurrentTime();
  clientInfoMutex.lock();
  if (count != NULL) *count = clientCount;
  if (delta != NULL) *delta = t - lastClientDisconnect;
  clientInfoMutex.unlock();
}

void ServerImpl::acceptClients(tcp::acceptor* acc)
{
  int clientId = 0;
  while (true)
  {
    tcp::socket* client = new tcp::socket(ioService);
    acc->accept(*client);
    clientId += 1;
    log << "Accepted client " << clientId << "\n";
    clientInfoMutex.lock();
    clientCount += 1;
    clientInfoMutex.unlock();
    client->set_option(ip::tcp::no_delay(true));
    client->set_option(socket_base::send_buffer_size(1048576));
    boost::thread serverThread(runClientThread, this, client, clientId);
  }
}

unsigned getInt(const unsigned char* buf)
{
  //Decode big-endian, from Java:
  return ((unsigned)(buf[0]) << 24) | ((unsigned)(buf[1]) << 16) | ((unsigned)(buf[2]) << 8) | ((unsigned)(buf[3]) << 0);
}

template <typename T>
void putInt(T* buf, const unsigned pos, const unsigned val)
{
  //Send big-endian, for Java:
  (*buf)[pos] = (val & 0xFF000000) >> 24;
  (*buf)[pos+1] = (val & 0x00FF0000) >> 16;
  (*buf)[pos+2] = (val & 0x0000FF00) >> 8;
  (*buf)[pos+3] = (val & 0x000000FF);
}

template <typename T>
void putShort(T* buf, const unsigned pos, const unsigned short val)
{
  //Send big-endian, for Java:
  (*buf)[pos] = (val & 0xFF00) >> 8;
  (*buf)[pos+1] = (val & 0x00FF);
}

void addInt(std::vector<unsigned char>* buf, const unsigned n)
{
  int s = buf->size();
  buf->resize(s + 4);
  putInt(buf, s, n);
}

void addFloat(std::vector<unsigned char>* buf, const float f)
{
  unsigned n;
  memcpy(&n,&f, 4);
  addInt(buf, n);
}

void addPoint3(std::vector<unsigned char>* buf, Point3 p)
{
  addFloat(buf, p.x);
  addFloat(buf, p.y);
  addFloat(buf, p.z);
}

//Firstly, an int saying how many Y lines to skip.  If pixels aren't present, this is 480.
//Then an int saying how many Y lines are present.  If pixels aren't present, this is 0.
//For each line, similarly an int with how many to skip followed by an int with how many are present, followed by ints for each.

// if the only pixel is 86 (0-based) then skipX will be 86, lastX will be 86
void calculateSkipLastX(unsigned short id, const unsigned short* labels, int* skipX, int* lastX)
{
      *skipX = 0;
      *lastX = 0;
      while (labels[*skipX] != id && *skipX < 640)
        {
          (*skipX) += 1;
        }
      if (*skipX < 640)
        {
          *lastX = 639;
          while (labels[*lastX] != id)
            {
              (*lastX) -= 1;
            }
        }

}

#define mymin(x,y) (x < y ? x : y)
#define mymax(x,y) (x > y ? x : y)

bool ServerImpl::getData(Data* buf)
{
  std::vector<unsigned char>* const v1 = &(buf->version1data);
  const int n = kinect->numUsers();
  if (n < 0)
    return false;
  v1->clear();
  v1->resize(4);
  v1->reserve(4 + 4 + n * sizeof(User));
  addInt(v1, n);
  const unsigned char* const rgb = kinect->getRGB();
  const unsigned short* const lbl = kinect->getUserLabels();
  const unsigned short* const dep = kinect->getDepth();
  if (rgb == NULL || lbl == NULL)
    return false;
  if (n > 0)
  {
    boost::scoped_array<User> users(new User[n]);
    if (false == kinect->getUsers(users.get()))
      return false;
    for (int i = 0; i < n; i++)
    {
      addInt(v1,users[i].id);
      addInt(v1,users[i].state);
      if (users[i].state == STATE_TRACKING)
      {
        for (int j = 0; j < NUM_JOINTS;j++)
        {
          addFloat(v1, users[i].joints[j].confidence);
          addPoint3(v1, users[i].joints[j].positionWorld);
          addPoint3(v1, users[i].joints[j].positionScreen);
        }
      }
    }
  }

  memcpy(buf->fullImageArray.c_array(), rgb, 640*480*3);  
  buf->maxDepth = kinect->getMaxDepth();
  memcpy(buf->fullDepthArray.c_array(), dep, 640*480*sizeof(unsigned short));  
  memcpy(buf->fullLabelArray.c_array(), lbl, 640*480*sizeof(unsigned short));

  putInt(v1, 0, v1->size() - 4);

  return true;
}

void sendSizeZero(tcp::socket* sock)
{
  std::vector<unsigned char> sizeZero;
  addInt(&sizeZero, 0);
  write(*sock, boost::asio::buffer(sizeZero));
}

void ServerImpl::handleConnection(tcp::socket* sock, int id)
{
  unsigned lastFrame = 0xFFFFFFFF; // won't match actual last frame
  unsigned char request[8];
  
  try
  {
    while (true)
    {
      read(*sock, boost::asio::buffer(request, 4));
      unsigned req = getInt(request);
      switch (req)
      {
        //      case 0:
      case 1:
      case 2:
      {
        read(*sock, boost::asio::buffer(request, 8));
        const unsigned width = getInt(request);
        const unsigned height = getInt(request + 4);

        const unsigned widthInc = 640 / width;
        const unsigned heightInc = 480 / height;

        if (width == 0 || height == 0 || (width <= 640 && height <= 480 && (640 % width) == 0 && (480 % height) == 0))
        {
          boost::shared_ptr< Data > dataToSend;
          do
          {
            //Wait for the monitor to signal the next item of data being ready:
            { // Must scope the lock properly:
              boost::unique_lock<boost::mutex> lock(dataNotifyMutex);
              while (dataNotifyFrame == lastFrame)
              {
                dataNotifyCond.wait(lock);
              }
              lastFrame = dataNotifyFrame;
            }

            //When it is ready, we still need the read lock before accessing:
            dataMutex.lock();
            dataToSend = sharedData;
            dataMutex.unlock();
          }
          while (dataToSend.get() == NULL);

          write(*sock, boost::asio::buffer(dataToSend->version1data));
          if (width != 0 && height != 0)
          {
            if (req == 1 || req == 2)
            {
              //compress and send the user labels:
              std::vector<unsigned char> rleLabels;
              const boost::array<unsigned short, 640*480>& labels(dataToSend->fullLabelArray);
              rleLabels.resize(4 + 640*480);
              unsigned short prevValue = 0x7FFF;
              unsigned short curCount = 0;
              unsigned dest = 4;
              unsigned i = 0;
              for (int y = 0; y < 480; y += heightInc, i += heightInc*640)
              {
                for (int x = 0; x < 640; x += widthInc)
                {
                  const unsigned short curValue = labels[i + x];
                  if (curValue == prevValue && curCount < 0x7FFF)
                  {
                    curCount += 1;
                  }
                  else
                  {
                    if (curCount == 1)
                    {
                      putShort(&rleLabels, dest, prevValue);
                      dest += sizeof(unsigned short);
                    }
                    else if (curCount > 1)
                    {
                      putShort(&rleLabels, dest, 0x8000 | curCount);
                      putShort(&rleLabels, dest + sizeof(unsigned short), prevValue);
                      dest += 2*sizeof(unsigned short);
                    }
                    prevValue = curValue;
                    curCount = 1;
                  }
                }
              }
              if (curCount == 1)
              {
                putShort(&rleLabels, dest, prevValue);
                dest += sizeof(unsigned short);
              }
              else if (curCount > 1)
              {
                putShort(&rleLabels, dest, 0x8000 | curCount);
                putShort(&rleLabels, dest + sizeof(unsigned short), prevValue);
                dest += 2*sizeof(unsigned short);
              }
              rleLabels.resize(dest);
              putInt(&rleLabels, 0, dest - 4);
              write(*sock, boost::asio::buffer(rleLabels));
            }

            //process and send the RGB image:
            boost::scoped_array<unsigned char> img(new unsigned char[width*height*4]);
            const boost::array<unsigned char, 640*480*3>& rgb(dataToSend->fullImageArray);
            unsigned destIndex = 0;
            unsigned i = 0;
            for (int y = 0; y < 480; y += heightInc, i += heightInc*640)
            {
              for (int x = 0; x < 640; x += widthInc)
              {
                if (i + x == 0)
                {
                  img[destIndex] = 0xFF;
                  memcpy(&(img[destIndex + 1]), &(rgb[(i+x)*3]), 3);
                }
                else
                {
                  //4-byte copy must be faster than 3:
                  memcpy(&(img[destIndex]), &(rgb[(i+x)*3 - 1]), 4);
                  img[destIndex] = 0xFF;
                }
                destIndex += 4;
              }
            }
            write(*sock, boost::asio::buffer(img.get(), width*height*4));
          } // else no need to send anything if they wanted a zero-sized image

          if (req == 2)
          {
            boost::scoped_array<unsigned char> depth(new unsigned char[(1 + 640*480)*sizeof(unsigned short)]);
            putShort(&depth, 0, dataToSend->maxDepth);
            for (int y = 0; y < 480; y++)
            {
              for (int x = 0; x < 640; x++)
              {
                putShort(&depth, (1 + y*640 + x)*sizeof(unsigned short), dataToSend->fullDepthArray[y*640 + x]);
              }
            }
 
            write(*sock, boost::asio::buffer(depth.get(), (1 + 640*480)*sizeof(unsigned short)));
          }
        }
        else // They asked for an image size that isn't an exact division of 640 * 480, ignore it
        {
          if (width != 0 && height != 0)
            log << "Client requested invalid size: " << width << " x " << height << "\n";
          sendSizeZero(sock);
        }
      }
      break;
      default:
      {
        sendSizeZero(sock);
      }
      break;
      }
    }
  }
  catch(...)
  {
    log << "Client gone " << id << "\n";
    clientInfoMutex.lock();
    clientCount -= 1;
    lastClientDisconnect.setCurrentTime();
    clientInfoMutex.unlock();
    //And finish...
  }
}


void writePPM(std::string path, const boost::array<unsigned char, 640*480*3 >& rgb)
{
  std::string header("P6 640 480 255\n");

  std::ofstream file;
  file.open(path.c_str(), std::ios::trunc | std::ios::binary);
  file.write(header.c_str(), header.size());
  for (int i = 0; i < 480; i++)
  {
    file.write((const char*)&(rgb[i* 640*3]), 640*3);
    //file.write("\n", 1);
  }
  file.close();
}

// Frequency in seconds:
#define IMAGE_FREQ 1
// Number to keep before overwriting them:
#define NUM_IMAGES 300

void ServerImpl::storeFeed()
{
  unsigned lastFrame = 0xFFFFFFFF; // won't match actual last frame
  unsigned imageIndex = 0;

  while (true)
  {
    //delay here until the right time:
    boost::this_thread::sleep(boost::posix_time::seconds(IMAGE_FREQ));

    //Wait for the monitor to signal the next item of data being ready:
    { // Must scope the lock properly:
      boost::unique_lock<boost::mutex> lock(dataNotifyMutex);
      while (dataNotifyFrame == lastFrame)
      {
        dataNotifyCond.wait(lock);
      }
      lastFrame = dataNotifyFrame;
    }

    //When it is ready, we still need the read lock before accessing:
    dataMutex.lock();
    boost::shared_ptr< Data > data(sharedData);
    dataMutex.unlock();

    if (data.get() != NULL)
    {
      char filename[1024];
      sprintf(filename, "%s/feed%.4d.ppm", feedDir.c_str(), imageIndex);
      imageIndex = (imageIndex + 1) % NUM_IMAGES;

      writePPM(filename, data->fullImageArray);
    }
  }
}
