/*
 * kinectserver - a program that send data from the Kinect sensor to connected clients.
 * Copyright (c) 2011, University of Kent.
 * Contact: Neil Brown, nccb@kent.ac.uk
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "interface.h"
#include "server.h"

#include <string.h>
#include <stdlib.h>

#include <boost/program_options.hpp>
#include <boost/smart_ptr.hpp>

namespace popt = boost::program_options;

bool setup(int argc, wchar_t** argv, Log log, Server* server, boost::scoped_ptr<Time>* idleTimeout)
{
  char configPath[1024] = "";
  bool matchDepthRGB = true;
  bool network = false;
  bool time = false;

  popt::options_description opts("Options");
  opts.add_options()
    ("config,c", popt::value<std::string>(), "supply custom config file")
    ("help,h", "show this help")
    ("idle,i", popt::value<unsigned>(), "set the idle timeout (in minutes): 0 disables idle state")
    ("matchoff,m", "suppress matching of the RGB and depth images (can aid performance)")
    ("network,n", "allow connections from remote machines across the network")
    ("security,s", popt::value<std::string>(), "keep latest images feed in given directory")
    ("time,t", "print frame timings to the log (warning: can produce a lot of output)")
  ;
  popt::variables_map vm;
  popt::store(popt::parse_command_line(argc, argv, opts), vm);
  popt::notify(vm);

  if (vm.count("help") > 0)
  {
    std::cout << opts << std::endl;
    return false;
  }

  matchDepthRGB = vm.count("matchoff") == 0;
  network = vm.count("network") > 0;
  time = vm.count("time") > 0;
  if (vm.count("config") > 0)
  {
    strcpy(configPath, vm["config"].as<std::string>().c_str());
  }

  if (vm.count("idle") > 0)
    {
      unsigned val = vm["idle"].as<unsigned>();
      if (val != 0)
        idleTimeout->reset(new Time(Time::minutes(val)));
      else
        idleTimeout->reset(NULL);
    }
  else
    idleTimeout->reset(new Time(Time::minutes(15))); 

  if (0 != strcmp(configPath, ""))
    {
      //They gave a config file; use it:
      log << "Using config file \"" << configPath << "\"...\n";
    }
  else
    {
      log << "Using default config (use -c to supply a config XML file)...\n";
    }

  if (matchDepthRGB)
    {
      log << "Correlating RGB with depth view, which can make server unstable on low-spec machines (use -m to turn off)\n";
    }
  else
    {
      log << "Not correlating RGB with depth view\n";
    }


  if (time)
    {
      log << "Displaying timings\n";
    }
  else
    {
      log << "Not displaying timings (use -t to print frame timings)\n";
    }

  if (network)
    {
      log << "Allowing anyone on your network to connect and access your feed\n";
    }
  else
    {
      log << "Only allowing connections from this machine (use -n to allow connections from remote machines)\n";
    }

  std::string securityDir = vm.count("security") > 0 ? vm["security"].as<std::string>() : "";

  return server->init(new KinectInterfaceFactory(log, configPath, matchDepthRGB, time), securityDir, network);
}
